import { Mapper } from '@automapper/core';
import { InjectMapper } from '@automapper/nestjs';
import { Controller, UsePipes } from '@nestjs/common';
import { CommandBus, QueryBus } from '@nestjs/cqrs';
import { GrpcMethod, RpcException } from '@nestjs/microservices';
import { RpcValidationPipe } from '../../../../utils/pipes/rpc.validation-pipe';
import { FindAdByUuidRequest } from '../../domain/dtos/find-ad-by-uuid.request';
import { AdPresenter } from './ad.presenter';
import { FindAdByUuidQuery } from '../../queries/find-ad-by-uuid.query';
import { Ad } from '../../domain/entities/ad';
import { CreateAdRequest } from '../../domain/dtos/create-ad.request';
import { CreateAdCommand } from '../../commands/create-ad.command';
import { DatabaseException } from '../../../database/exceptions/database.exception';

@UsePipes(
  new RpcValidationPipe({
    whitelist: false,
    forbidUnknownValues: false,
  }),
)
@Controller()
export class AdController {
  constructor(
    private readonly _commandBus: CommandBus,
    private readonly queryBus: QueryBus,
    @InjectMapper() private readonly _mapper: Mapper,
  ) {}

  @GrpcMethod('AdsService', 'FindOneByUuid')
  async findOnebyUuid(data: FindAdByUuidRequest): Promise<AdPresenter> {
    try {
      const ad = await this.queryBus.execute(new FindAdByUuidQuery(data));
      return this._mapper.map(ad, Ad, AdPresenter);
    } catch (e) {
      throw new RpcException({
        code: e.code,
        message: e.message,
      });
    }
  }

  @GrpcMethod('AdsService', 'Create')
  async createAd(data: CreateAdRequest): Promise<AdPresenter> {
    try {
      const ad = await this._commandBus.execute(new CreateAdCommand(data));
      return this._mapper.map(ad, Ad, AdPresenter);
    } catch (e) {
      if (e instanceof DatabaseException) {
        if (e.message.includes('Already exists')) {
          throw new RpcException({
            code: 6,
            message: 'Ad already exists',
          });
        }
      }
      throw new RpcException({});
    }
  }
}
