import { Test, TestingModule } from '@nestjs/testing';
import { CreateAdUseCase } from '../../domain/usecases/create-ad.usecase';
import { CreateAdRequest } from '../../domain/dtos/create-ad.request';
import { Messager } from '../../adapters/secondaries/messager';
import { AdsRepository } from '../../adapters/secondaries/ads.repository';
import { CreateAdCommand } from '../../commands/create-ad.command';
import { AutomapperModule } from '@automapper/nestjs';
import { classes } from '@automapper/classes';
import { Frequency } from '../../domain/types/frequency.enum';
import { Ad } from '../../domain/entities/ad';
import { AdProfile } from '../../mappers/ad.profile';
import { AddressRequestDTO } from '../../domain/dtos/create.address.request';
import { AdCreation } from '../../domain/dtos/ad.creation';
import { AddressCreation } from '../../domain/dtos/address.creation';

const mockAddress1: AddressRequestDTO = {
  position: 0,
  lon: 48.68944505415954,
  lat: 6.176510296462267,
  houseNumber: '5',
  street: 'Avenue Foch',
  locality: 'Nancy',
  postalCode: '54000',
  country: 'France',
};
const mockAddress2: AddressRequestDTO = {
  position: 1,
  lon: 48.8566,
  lat: 2.3522,
  locality: 'Paris',
  postalCode: '75000',
  country: 'France',
};
const mockAddressWithoutPos1: AddressRequestDTO = {
  lon: 43.2965,
  lat: 5.3698,
  locality: 'Marseille',
  postalCode: '13000',
  country: 'France',
};
const mockAddressWithoutPos2: AddressRequestDTO = {
  lon: 43.7102,
  lat: 7.262,
  locality: 'Marseille',
  postalCode: '06000',
  country: 'France',
};
const minimalReccurentAdREquest: CreateAdRequest = {
  userUuid: '224e0000-0000-4000-a000-000000000000',
  frequency: Frequency.RECURRENT,
  fromDate: new Date('01-05-2023'),
  toDate: new Date('01-05-2024'),
  schedule: {
    mon: '08:00',
  },
  marginDurations: {},
  addresses: [mockAddress1, mockAddress2],
};
const newAdRequest: CreateAdRequest = {
  userUuid: '113e0000-0000-4000-a000-000000000000',
  driver: true,
  passenger: false,
  frequency: Frequency.RECURRENT,
  fromDate: new Date('01-05-2023'),
  toDate: new Date('20-08-2023'),
  schedule: {
    tue: '08:00',
    wed: '08:30',
  },
  marginDurations: {
    mon: undefined,
    tue: undefined,
    wed: undefined,
    thu: undefined,
    fri: undefined,
    sat: undefined,
    sun: undefined,
  },
  seatsDriver: 2,
  addresses: [mockAddress1, mockAddress2],
};

const mockMessager = {
  publish: jest.fn().mockImplementation(),
};
const mockDefaultParamsProvider = {
  getParams: () => {
    return {
      MON_MARGIN: 900,
      TUE_MARGIN: 900,
      WED_MARGIN: 900,
      THU_MARGIN: 900,
      FRI_MARGIN: 900,
      SAT_MARGIN: 900,
      SUN_MARGIN: 900,
      DRIVER: false,
      SEATS_PROVIDED: 0,
      PASSENGER: true,
      SEATS_REQUESTED: 1,
      STRICT: false,
    };
  },
};
const mockAdRepository = {
  create: jest
    .fn()
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    .mockImplementationOnce((command?: CreateAdCommand) => {
      return Promise.resolve({
        ...newAdRequest,
        uuid: 'ad000000-0000-4000-a000-000000000000',
        createdAt: new Date('01-05-2023'),
      });
    })
    .mockImplementationOnce(() => {
      throw new Error('Already exists');
    })
    .mockImplementation(),
};
describe('CreateAdUseCase', () => {
  let createAdUseCase: CreateAdUseCase;
  beforeAll(async () => {
    const module: TestingModule = await Test.createTestingModule({
      imports: [AutomapperModule.forRoot({ strategyInitializer: classes() })],
      providers: [
        {
          provide: AdsRepository,
          useValue: mockAdRepository,
        },
        {
          provide: Messager,
          useValue: mockMessager,
        },
        CreateAdUseCase,
        AdProfile,
        {
          provide: 'ParamsProvider',
          useValue: mockDefaultParamsProvider,
        },
      ],
    }).compile();

    createAdUseCase = module.get<CreateAdUseCase>(CreateAdUseCase);
  });
  it('should be defined', () => {
    expect(createAdUseCase).toBeDefined();
  });
  describe('execution', () => {
    const newAdCommand = new CreateAdCommand(newAdRequest);
    it('should create an new ad', async () => {
      const newAd: Ad = await createAdUseCase.execute(newAdCommand);
      expect(newAd.userUuid).toBe(newAdRequest.userUuid);
      expect(newAd.uuid).toBeDefined();
    });
    it('should throw an error if the ad already exists', async () => {
      await expect(
        createAdUseCase.execute(newAdCommand),
      ).rejects.toBeInstanceOf(Error);
    });
  });

  describe('Ad parameter default setting ', () => {
    beforeEach(() => {
      mockAdRepository.create.mockClear();
    });

    it('should define mimimal ad as 1 passager add', async () => {
      const newAdCommand = new CreateAdCommand(minimalReccurentAdREquest);
      await createAdUseCase.execute(newAdCommand);
      const expectedAdCreation = {
        userUuid: minimalReccurentAdREquest.userUuid,
        frequency: minimalReccurentAdREquest.frequency,
        fromDate: minimalReccurentAdREquest.fromDate,
        toDate: minimalReccurentAdREquest.toDate,
        monTime: minimalReccurentAdREquest.schedule.mon,
        tueTime: undefined,
        wedTime: undefined,
        thuTime: undefined,
        friTime: undefined,
        satTime: undefined,
        sunTime: undefined,
        monMargin: mockDefaultParamsProvider.getParams().MON_MARGIN,
        tueMargin: mockDefaultParamsProvider.getParams().TUE_MARGIN,
        wedMargin: mockDefaultParamsProvider.getParams().WED_MARGIN,
        thuMargin: mockDefaultParamsProvider.getParams().THU_MARGIN,
        friMargin: mockDefaultParamsProvider.getParams().FRI_MARGIN,
        satMargin: mockDefaultParamsProvider.getParams().SAT_MARGIN,
        sunMargin: mockDefaultParamsProvider.getParams().SUN_MARGIN,
        driver: mockDefaultParamsProvider.getParams().DRIVER,
        seatsDriver: mockDefaultParamsProvider.getParams().SEATS_PROVIDED,
        passenger: mockDefaultParamsProvider.getParams().PASSENGER,
        seatsPassenger: mockDefaultParamsProvider.getParams().SEATS_REQUESTED,
        strict: mockDefaultParamsProvider.getParams().STRICT,
        addresses: {
          create: minimalReccurentAdREquest.addresses as AddressCreation[],
        },
        createdAt: undefined,
      } as AdCreation;

      expect(mockAdRepository.create).toBeCalledWith(expectedAdCreation);
    });
    it('should create an passengerAd with addresses without position ', async () => {
      const newPunctualPassengerAdRequest: CreateAdRequest = {
        userUuid: '113e0000-0000-4000-a000-000000000000',
        passenger: true,
        frequency: Frequency.PUNCTUAL,
        departure: new Date('05-22-2023 09:36'),

        marginDurations: {
          mon: undefined,
          tue: undefined,
          wed: undefined,
          thu: undefined,
          fri: undefined,
          sat: undefined,
          sun: undefined,
        },
        seatsPassenger: 1,
        addresses: [mockAddressWithoutPos1, mockAddressWithoutPos2],
        schedule: {},
      };
      const newAdCommand = new CreateAdCommand(newPunctualPassengerAdRequest);
      await createAdUseCase.execute(newAdCommand);
      const expectedAdCreation = {
        userUuid: newPunctualPassengerAdRequest.userUuid,
        frequency: newPunctualPassengerAdRequest.frequency,
        fromDate: newPunctualPassengerAdRequest.departure,
        toDate: newPunctualPassengerAdRequest.departure,
        monTime: '09:36',
        tueTime: undefined,
        wedTime: undefined,
        thuTime: undefined,
        friTime: undefined,
        satTime: undefined,
        sunTime: undefined,
        monMargin: mockDefaultParamsProvider.getParams().MON_MARGIN,
        tueMargin: mockDefaultParamsProvider.getParams().TUE_MARGIN,
        wedMargin: mockDefaultParamsProvider.getParams().WED_MARGIN,
        thuMargin: mockDefaultParamsProvider.getParams().THU_MARGIN,
        friMargin: mockDefaultParamsProvider.getParams().FRI_MARGIN,
        satMargin: mockDefaultParamsProvider.getParams().SAT_MARGIN,
        sunMargin: mockDefaultParamsProvider.getParams().SUN_MARGIN,
        driver: false,
        seatsDriver: 0,
        passenger: newPunctualPassengerAdRequest.passenger,
        seatsPassenger: newPunctualPassengerAdRequest.seatsPassenger,
        strict: mockDefaultParamsProvider.getParams().STRICT,
        addresses: {
          create: newPunctualPassengerAdRequest.addresses as AddressCreation[],
        },
        createdAt: undefined,
      } as AdCreation;
      expect(mockAdRepository.create).toBeCalledWith(expectedAdCreation);
    });
  });
});
